/*
 * CipherInputStream.java
 *
 * This class is part of our JCE 1.2 cleanroom implementation
 *
 * Copyright (c) 2000 Virtual Unlimited B.V.
 *
 * Author: Bob Deblier <bob@virtualunlimited.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package javax.crypto;

import java.io.*;

/* VERY IMPORTANT NOTE: 
 * The CipherInputStream should keep at least the cipher's blockSize bytes in
 * filterBuffer until the filter InputStream is at end-of-stream.
 * Otherwise decryption/unpadding will not work. Neither Sun nor Cryptix
 * handle this properly.
 */
public class CipherInputStream extends FilterInputStream
{
	private Cipher cipher;
	private InputStream filter;

	private byte[] filterBuffer;
	private byte[] updateBuffer;

	private int blockSize;
	private int filterOffset;
	private int updateOffset;
	private int updateLength;

	private boolean eos;

	protected CipherInputStream(InputStream is)
	{
		this(is, new NullCipher());
	}

	public CipherInputStream(InputStream is, Cipher c)
	{
		super(is);
		cipher = c;
		filter = is;
		blockSize = c.getBlockSize();
		filterBuffer = new byte[1024 + blockSize];
		filterOffset = 0;
		updateBuffer = null;
		updateOffset = 0;
		updateLength = 0;
		eos = false;
	}

	public int available()
	{
		return updateLength - updateOffset;
	}

	public void close() throws IOException
	{
		if (!eos)
		{
			try
			{
				cipher.doFinal();
			}
			catch (BadPaddingException e)
			{
			}
			catch (IllegalBlockSizeException e)
			{
			}
		}
		filter.close();
	}

	public boolean markSupported()
	{
		return false;
	}

	private int bufferedRead() throws IOException
	{
		if (eos)
			return -1;

		while (filterOffset < filterBuffer.length)
		{
			int rc = filter.read(filterBuffer, filterOffset, filterBuffer.length - filterOffset);

			if (rc == -1)
			{
				eos = true;
				updateOffset = 0;
				try
				{
					updateBuffer = cipher.doFinal(filterBuffer, 0, filterOffset);
				}
				catch (IllegalBlockSizeException e)
				{
					updateBuffer = null;
				}
				catch (BadPaddingException e)
				{
					updateBuffer = null;
				}

				return updateLength = (updateBuffer == null) ? 0 : updateBuffer.length;
			}
			else
			{
				filterOffset += rc;
			}
		}

		updateOffset = 0;
		try
		{
			/* update everything except the last blockSize bytes */
			updateBuffer = cipher.update(filterBuffer, 0, filterBuffer.length - blockSize);
			System.arraycopy(filterBuffer, filterBuffer.length - blockSize, filterBuffer, 0, blockSize);
			filterOffset = blockSize;
		}
		catch (IllegalStateException e)
		{
			updateBuffer = null;
		}

		return updateLength = (updateBuffer == null) ? 0 : updateBuffer.length;
	}

	public int read() throws IOException
	{
		if (updateOffset >= updateLength)
		{
			int rc = 0;
			while (rc == 0)
				rc = bufferedRead();
			if (rc == -1)
				return -1;
		}
		return (updateBuffer[updateOffset++] & 0xff);
	}

	public int read(byte[] b) throws IOException
	{
		return read(b, 0, b.length);
	}
	
	public int read(byte[] b, int off, int len) throws IOException
	{
		if (updateOffset >= updateLength)
		{
			int rc = 0;
			while (rc == 0)
				rc = bufferedRead();

			if (rc == -1)
				return -1;
		}

		int avail = updateLength - updateOffset;

		if (len <= 0 || avail <= 0)
			return 0;

		if (len > avail)
			len = avail;

		System.arraycopy(updateBuffer, updateOffset, b, off, len);

		updateOffset += len;

		return len;
	}

	public long skip(long n) throws IOException
	{
		if (n < 0L)
			return 0L; 

		int avail = updateLength - updateOffset;

		if (n > (long) avail)
			n = (long) avail;
			
		updateOffset += (int) n;
		return n;
	}
}
